/* -----------------------------------------------------------	*/
/* ShadowVolume Tutorial										*/
/* -----------------------------------------------------------	*/
/* by n0rmAnfreAK												*/
/* visit www.normanfreak.de.vu for updates						*/
/*																*/
/* this current version is pretty hardcoded						*/
/* but it should show the general idea behind shadow volumes	*/
/* and the use of edgelists for faster generation				*/
/*																*/
/* feel free to email me any critic and comments				*/
/* ------------------------------------------------------------ */

#include <windows.h>
#include <gl\glut.h>
#include <stdio.h>
#include <math.h>

#include "Edge.h"
#include "SVHelper.h"

#include <list>

using namespace std;

//float lightx, lighty, lightz;
float light[2][3];
bool lightOn[2];
float lightchange;
float alpha;
int frames;
long time, lastTime, animationTime;
char title[100];


// here is the geometry defined which will cast the shadow
// define the positions of a box with edgelength = .5
float boxv [][3] =	{
	{0, 1, .5},
	{1, 1, .5},
	{1, 1, 1.5},
	{0, 1, 1.5},
	{0, 2, .5},
	{1, 2, .5},
	{1, 2, 1.5},
	{0, 2, 1.5}
};

// define the 6 boxplanes (indices to points)
int boxplanes [6][4] = {
	{0, 3, 2, 1},
	{4, 5, 6, 7},
	{0, 1, 5, 4},
	{1, 2, 6, 5},
	{2, 3, 7, 6},
	{3, 0, 4, 7}
};

// define the normals for the sides
float boxnormals [6][3] = {
	{0, -1 ,0},
	{0, 1, 0},
	{0, 0, -1},
	{1, 0, 0},
	{0, 0, 1},
	{-1,0, 1}
};

// generate a list of all edges the geometry has
list<Edge*> generateEdgelist(int *faces, int nbrFaces, int type)
{
	list<Edge*> edgelist;
	int a,b;
	bool found;
	
	// go through faces
	for(int i=0; i<nbrFaces; i++)
	{
		// check the edges
		for(int s=0; s<type; s++)
		{
			a = faces[i*type+s];
			b = faces[i*type+((s+1)%type)];

			found = false;

			// iterate through current edgelist and check if edge already exists			
			if(!edgelist.empty() )
			{
				for(std::list<Edge*>::iterator it = edgelist.begin(); it != edgelist.end(); ++it)
				{
					if( (*it)->equals(a, b))
					{
						(*it)->faceBpoint = faces[i*type+(s+2)%type];	// just store the third point to define the face
						found = true;
						break;
					}
				}
			}
			if(!found)
			{
				edgelist.push_back(new Edge(faces[i*type+(s+2)%type], a, b) );
			}
		}
	}

	return edgelist;
}

list<Edge*> mylist;

// all code that has to be run once before the rendering
void Init()	
{
	glClearColor(0.33f,0.225f,0.0f,1.0f);
	glClearStencil(0);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_CULL_FACE);
	glClearDepth(1.0);

	light[0][0] = -4;
	light[0][1] = 7;
	light[0][2] = 4.5;
	lightOn[0] = true;

	light[1][0] = -4;
	light[1][1] = 8;
	light[1][2] = 4;
	lightOn[1] = true;
	
	lightchange = 0.2;

	alpha = 0;
	frames = 0;
	time = 0;
	lastTime = animationTime = GetTickCount();	
	int *tmp = &boxplanes[0][0];

	float color[4];
	color[0] = .8f;
	color[1] = .2f;
	color[2] = .2f;
	color[3] = 1.f;
	glLightfv(GL_LIGHT0, GL_DIFFUSE, color);
	color[0] = .1f;
	color[1] = .1f;
	color[2] = .9f;
	glLightfv(GL_LIGHT1, GL_DIFFUSE, color);

	// generate our edgelist ....shouldnt be stored globally in a nice program ;)
	mylist = generateEdgelist(tmp, 6, 4);
	printf(" listlength is : %i\n", mylist.size());
	
	for(std::list<Edge*>::iterator it = mylist.begin(); it != mylist.end(); ++it)
		(*it)->print();
}


void makeShadowVolumes(int lightindex)
{
	glBegin(GL_QUADS);
	glColor3f(1.f, 0.f, 0.f);
	
	bool face;
	float a[3], b[3];
	float *pnts = boxv[0];

	// go through all edges 
	for(std::list<Edge*>::iterator i = mylist.begin(); i != mylist.end(); ++i)
	{
		// did we find a contouredge
		if(SVHelper::isContourEdge( (*i), pnts, light[lightindex], face) )
		{
			int indexA = (*i)->pointA;
			int indexB = (*i)->pointB;
			a[0] = pnts[indexA*3];
			a[1] = pnts[indexA*3+1];
			a[2] = pnts[indexA*3+2];
			b[0] = pnts[indexB*3];
			b[1] = pnts[indexB*3+1];
			b[2] = pnts[indexB*3+2];
			
			// if we find a contouredge stretch it to create a shadowvolume  
			if(face)
				SVHelper::stretchEdge(a, b, light[lightindex]);
			else
				SVHelper::stretchEdge(b, a, light[lightindex]);
		}
	}
	glEnd();
}

void Scene();
void ShadowRun()
{	
	for(int i=0; i<2; i++)
	{
		if(!lightOn[i])
			continue;
		glClear(GL_STENCIL_BUFFER_BIT);

		// turn off visible rendering 
		glColorMask( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
		// don't override depthvalues
		glDepthMask( GL_FALSE);
	
		glStencilFunc(GL_ALWAYS, ~0, 0);
		glEnable(GL_STENCIL_TEST);
		
		// count frontfacing shadowvolumepolygons		
		// how many times do we enter a shadowvolume
		glCullFace(GL_BACK);
		glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
		makeShadowVolumes(i);

		// count backfacing shadowvolumepolygons
		// how many times do we leave a shadowvolume
		glCullFace(GL_FRONT);
		glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);	
		makeShadowVolumes(i);
		
		// turn on visible rendering
		glColorMask( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
		glDepthMask( GL_TRUE);

		// setup light
		switch(i)
		{
		case 0:
			glLightfv(GL_LIGHT0, GL_POSITION, light[0]);
			glEnable(GL_LIGHT0);
			break;
		case 1:
			glLightfv(GL_LIGHT1, GL_POSITION, light[1]);
			glEnable(GL_LIGHT1);
			break;
		}
		
		glCullFace(GL_BACK);
		glDepthFunc(GL_EQUAL);

		// test for fragments not in shadow
		// if we stepped out of the volume as often as we stepped in 
		// we are not inside a shadowvolume
		glStencilFunc(GL_EQUAL, 0, 0xff);

		// now render the scene with light...just the fragments not in shadow
		Scene();
	
		glDisable(GL_STENCIL_TEST);
		glDepthFunc(GL_LEQUAL);
	}
	glDisable(GL_LIGHT0);
	glDisable(GL_LIGHT1);	
}

void Scene()
{
	
	glLoadIdentity ();
	gluLookAt(0, 3, -3, 0, 0, 1, 0, 1, 0); 
	glRotatef ( 20.0, 1.0, 0.0, 0.0 );
	glScalef (0.5, 0.5, 0.5 );

	// draw ground
	glColor3f(.4, .4, .4);
	glFrontFace(GL_CW);
	glBegin(GL_QUADS);
	glNormal3f(0, 1, 0);
	glVertex3f(-50, -1.5, -50);
	glVertex3f(50, -1.5, -50);
	glVertex3f(50, -1.5, 50);
	glVertex3f(-50, -1.5, 50);
	glEnd();
	
	int a;

	// draw box
	glBegin(GL_QUADS);
	// go through boxsides
	for(a=0; a<6; a++)
	{
		// draw each side
		glNormal3f(boxnormals[a][0], boxnormals[a][1], boxnormals[a][2]);
		for(int b=0; b<4; b++)
			glVertex3f(boxv[boxplanes[a][b]][0], boxv[boxplanes[a][b]][1], boxv[boxplanes[a][b]][2]);
	}

	glEnd();
	glFrontFace(GL_CCW);	
}


void RenderScene(void)
{
	// clear all buffers
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

	// render Scene ambient	
	Scene();
	// render Scene light with shadowTest
	glBlendFunc(GL_ONE, GL_ONE);
	glEnable(GL_BLEND);
	ShadowRun();	
	glDisable(GL_BLEND);

	glutSwapBuffers (); 
	glFlush ();

	// bruteforce to get fps-count
	frames++;
	long time = GetTickCount();
	if(time - lastTime > 1000)
	{
		lastTime = time;
		sprintf(title, "StencilShadowVolumes  v0.9 - fps : %i", frames);
		glutSetWindowTitle(title);
		frames = 0;
	}
}

// all code to reshape the window
void Reshape(int width,int height)
{	
	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity();
	glViewport(0,0,width,height);
	
	GLfloat fAspect;
	if(height==0)
		height=1;
	fAspect=(float)width/(float)height;

	gluPerspective(60.0f, fAspect, 0.1, 30.0);
	glMatrixMode(GL_MODELVIEW); 
}

// all code to animate the scene is placed here
void Animate ()
{
	long time = GetTickCount();
	if(time - animationTime > 10)
	{
		// rotate the 2. light
		alpha += .02;
		if(alpha > 6.28)
			alpha -= 6.28;

		// move the 1. light up and down
		light[0][1] += lightchange;
		if(light[0][1] > 10)
			lightchange = -.2;
		else if(light[0][1] < 4)
			lightchange = .2;
		light[1][0] = 5 * cos(alpha);
		light[1][2] = 5 * sin(alpha);
		animationTime = time;
	}
	RenderScene();
}



void Keyboard(unsigned char key, int x, int y)
{
	// perform keypressed here
	switch(key)
	{
	case '1':
		lightOn[0] = !lightOn[0];
		break;

	case '2':
		lightOn[1] = !lightOn[1];
		break;
	}
	printf(" light0 is %i, light1 is %i\n", lightOn[0], lightOn[1]);
}

int main(int argc, char **argv)
{
	printf ("ShadowVolume Tutorial \n\n");
	glutInit ( &argc, argv );
	glutInitDisplayMode ( GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH|GLUT_STENCIL );
	glutInitWindowSize ( 600,600 );	
	glutCreateWindow ("StencilShadowVolumes  v0.9");
	glutDisplayFunc ( RenderScene );
	glutReshapeFunc ( Reshape );
	glutIdleFunc ( Animate );
	glutKeyboardFunc ( Keyboard );
	Init();
	glutMainLoop ();
	return 0;
}
